
/*
 * Copyright 2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Diorite
{

public delegate void UriOpener(string uri);

/**
 * Simple Document View is supposed to properly display content of SimpleDocBuffer.
 * 
 * It understands all formating tags of SimpleDocBuffer, sets proper link color and
 * manages link actions (changes cursor and emits link_clicked signal).
 */
public class SimpleDocView: Gtk.TextView
{
	[Description(nick = "Link opener", blurb = "Function to be used to open links.")]
	public UriOpener? link_opener {get; owned set;}
	[Description(nick = "Image opener", blurb = "Function to be used to open images.")]
	public UriOpener? image_opener {get; owned set;}
	private Gdk.Cursor? cursor = null;
	
	/**
	 * Creates new SimpleDocView.
	 */
	public SimpleDocView(SimpleDocBuffer? buffer=null)
	{
		Object(editable: false);
		link_opener = Diorite.System.open_uri_with_fallback;
		image_opener = Diorite.System.open_uri_with_fallback;
		this.buffer = buffer ?? new SimpleDocBuffer();
	}
	
	/**
	 * Emitted when a link is clicked. The default handler opens the link
	 * via link_opener field.
	 * 
	 * @param uri	target URI of the link
	 */
	public virtual signal void link_clicked(string uri)
	{
		debug("Open link: %s", uri);
		if (link_opener != null)
			link_opener(uri);
	}
	
	/**
	 * Emitted when an image is clicked. The default handler opens the image
	 * via image_opener field.
	 * 
	 * @param path	path of the image
	 */
	public virtual signal void image_clicked(string path)
	{
		debug("Open image: %s", path);
		if (image_opener != null)
			image_opener(path);
	}
	
	public override void realize()
	{
		base.realize();
		set_link_color();
	}
	
	public override void style_updated(){
		base.style_updated();
		if (get_realized())
			set_link_color();
	}
	
	private void set_link_color(){
		Gdk.RGBA? link_color = null;
		var doc_buffer = buffer as SimpleDocBuffer;
		if (doc_buffer != null
		&& get_style_context().lookup_color("link_color", out link_color))
		{
			doc_buffer.link_color = link_color;
		}
	}
	
	public override bool button_release_event(Gdk.EventButton event)
	{
		var cont = base.button_release_event(event);
		if (event.button == 1)
		{
			int x, y;
			window_to_buffer_coords(Gtk.TextWindowType.TEXT, (int) event.x, (int) event.y, out x, out y);
			unowned SimpleDocLink link;
			if (get_link_at_pos(x, y, out link))
			{
				link_clicked(link.uri);
			}
			else
			{
				var pixbuf = get_pixbuf_at_pos(x, y);
				if (pixbuf != null)
				{
					var path = pixbuf.get_data<string?>(SimpleDocBuffer.IMAGE_PATH);
					if (path != null)
					{
						image_clicked(path);
					}
				}
			}
		}
		return cont;
	}
	
	/**
	 * Get link at given position
	 * 
	 * @param x		x coordinate
	 * @param y		y coordinate
	 * @param link	link when found
	 * @return	true if a link has been found
	 */
	public bool get_link_at_pos(int x, int y, out unowned SimpleDocLink link)
	{
		link = null;
		Gtk.TextIter iter;
		get_iter_at_location(out iter, x, y);
		foreach(weak Gtk.TextTag tag in iter.get_tags())
		{
			weak SimpleDocLink? maybe_link = tag as SimpleDocLink;
			if (maybe_link != null)
			{
				link = maybe_link;
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Check whether position is inside iter area.
	 * 
	 * @param iter	iter
	 * @param x		x coordinate
	 * @param y		y coordinate
	 * @return	true if (x, y) is inside area of iter
	 */
	public bool is_in_iter_area(Gtk.TextIter iter, int x, int y)
	{
		Gdk.Rectangle area;
		get_iter_location(iter, out area);
		return x >= area.x && x <= area.x + area.width
		&& y >= area.y && y <= area.y + area.height;
	}
	
	/**
	 * Get pixbuf at given position.
	 * 
	 * @param x		coordinate x
	 * @param y		coordinate y
	 * @return pixbuf if found, null otherwise
	 */
	public Gdk.Pixbuf? get_pixbuf_at_pos(int x, int y)
	{
		Gtk.TextIter iter;
		get_iter_at_location(out iter, x, y);
		var pixbuf = iter.get_pixbuf();
		if (pixbuf != null && is_in_iter_area(iter, x, y))
			return pixbuf;
		
		/* When mouse cursor in over the second half of a pixbuf, iter on the right hand side
		 * is returned instead the right one, so we have to go backward and
		 * check a pixbuf presence again. */
		 
		iter.backward_char();
		pixbuf = iter.get_pixbuf();
		if (pixbuf != null && is_in_iter_area(iter, x, y))
			return pixbuf;
		
		return null;
	}
	
	public override bool motion_notify_event(Gdk.EventMotion event)
	{
		var cont = base.motion_notify_event(event);
		int x, y;
		window_to_buffer_coords(Gtk.TextWindowType.TEXT, (int) event.x, (int) event.y, out x, out y);
		update_cursor(x, y);
		return cont;
	}
	
	private void update_cursor(int x, int y)
	{
		var display = get_display();
		Gdk.Cursor cursor = get_link_at_pos(x, y, null) || get_pixbuf_at_pos(x, y) != null
			? new Gdk.Cursor.for_display(display, Gdk.CursorType.HAND2) : null;
		
		if (this.cursor != cursor)
		{
			get_window(Gtk.TextWindowType.TEXT).set_cursor(cursor);
			display.flush();
			this.cursor = cursor;
		}
	}
}

} // namespace Diorite
