/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Diorite{

	/**
	 * Callback for Actions
	 */
	public delegate void SimpleActionCallback();
	
	/**
	 * Class managing UI actions
	 */
	public class Actions: GLib.Object{
		private Gtk.UIManager ui_manager;
		private Gtk.ActionGroup actions;
		private string group_name;
		
		/**
		 * Creates new actions manager
		 * 
		 * @param ui_manager	UI Manager actions will be attached to
		 * @param group_name	Name of the default action group
		 */
		public Actions(Gtk.UIManager ui_manager, string group_name){
			this.ui_manager = ui_manager;
			this.group_name = group_name;
			actions = new Gtk.ActionGroup(group_name);
			ui_manager.insert_action_group(actions, 0);
		}
		
		public signal void action_changed(Action action, ParamSpec p);
		
		/**
		 * Adds new action.
		 * 
		 * @param action	Action to be added.
		 * @return 			true on success, false on failure (conflict)
		 */
		public bool add_action(Action action){
			if(!is_unique(action.name)) return false;
			add_action_no_check(action);
			return true;
		}
		
		/**
		 * Adds new action without checking its uniqueness. Make sure you used is_unique()
		 * before calling this method.
		 * 
		 * @param action	Action to be added.
		 */
		private void add_action_no_check(Action action){
//~ 			lib_debug("New action: %s", action.name);
			action.notify["keybinding"].connect_after(on_keybinding_changed);
			var accel_path = accel_path_for_name(action.name);
			action.set_accel_path(accel_path);
			actions.add_action(action);
			set_keybinding(accel_path, action.keybinding);
			action.activate.connect(debug_action);
			action.notify.connect_after(on_action_notify);
		}
		
		private void debug_action(Gtk.Action action){
			lib_debug("Action activated: %s", action.name);
		}
		
		/**
		 * Adds multiple new actions. Use only for NormalActions or ToggleActions,
		 * add_radio_actions() is more suitable for RadioActions.
		 * 
		 * @param actions	Actions to be added
		 * @return 			true on success, false on failure (conflict)
		 */
		public bool add_actions(Action[] actions){
			return_val_if_fail(actions.length > 0, false);
			var result = true;
			foreach(var action in actions){
				result = result && add_action(action);
			}
			return result;
		}
		
		/**
		 * Adds multiple radio actions and binds them to a group.
		 * 
		 * @param int		Current value within a group of radio options
		 * @param actions	Actions to be added
		 * @param on_change	callback on change
		 * @return 			true on success, false on failure (conflict)
		 */
		public bool add_radio_actions(RadioAction[] actions, int current_value){
			return_val_if_fail(actions.length > 0, false);
			var result = true;
			// Empty SList is actually NULL in C, we must use unowned to prevent Vala
			// from freeing the list
			unowned SList<Gtk.RadioAction> group = null;
			foreach(var action in actions){
				if(!is_unique(action.name)){
					result = false;
					continue;
				}
				action.set_group(group);
				group = action.get_group();
				action.active = current_value == action.value;
				add_action_no_check(action);
			}
			
			return result;
		}
		
		/**
		 * Retrieves action by name.
		 * 
		 * @param name	Action to be requested
		 * @return		Action instance or null if not found
		 */
		public unowned Action? get_action(string name){
			return actions.get_action(name) as Action;
		}
		
		/**
		 * Removes action by name
		 * 
		 * @param name	Action to be removed
		 * @return		true on success, false on failure
		 */
		public bool remove_action(string name){
			var action = get_action(name);
			if(action == null) return false;
			var accel_path = accel_path_for_name(name);
			Gtk.AccelMap.change_entry(accel_path, 0, 0, true);
			actions.remove_action(action);
			action.notify["keybinding"].disconnect(on_keybinding_changed);
			action.notify.disconnect(on_action_notify);
			return true;
		}
		
		/**
		 * Updates keybinding
		 */
		private void on_keybinding_changed(GLib.Object object, ParamSpec p){
			var action = object as Action;
			return_if_fail(action != null);
			set_keybinding(accel_path_for_name(action.name), action.keybinding);
		}
		
		/**
		 * Generates accelerator path for Action name
		 */
		private string accel_path_for_name(string name){
			return "<Actions>/%s/%s".printf(group_name, name);
		}
		
		private static void set_keybinding(string accel_path, string? keybinding){
			if(keybinding == null){
//~ 				lib_debug("Remove keybinding: %s", accel_path);
				Gtk.AccelMap.change_entry(accel_path, 0, 0, true);
			}
			else{
//~ 				lib_debug("Set keybinding: %s = %s", accel_path, keybinding);
				uint key;
				Gdk.ModifierType mods;
				Gtk.accelerator_parse(keybinding, out key, out mods);
				Gtk.AccelMap.change_entry(accel_path, key, mods, true);
			}
		}
		
		/**
		 * Returns true if action name is unique.
		 */
		private bool is_unique(string name){
			if(actions.get_action(name) == null) return true;
			
			lib_warning("Nuvola.Actions: Action %s cannot be added, because it already exists.", name);
			return false;
		}
		
		public bool trigger_action(string name){
			var action = get_action(name);
			if(action == null) return false;
			action.activate();
			return true;
		}
		
		private void on_action_notify(GLib.Object o, ParamSpec p)
		{
			var action = o as Action;
			if (action != null)
				action_changed(action, p);
		}
	}
	
	/**
	 * Common interface for all actions, that can be attached to user interface
	 * (a menu or tool bar item, systray, jumplists).
	 */
	public interface Action: Gtk.Action{
		[Description(nick = "Local keybinding", blurb = "Keyboard shortcut accessible inside window.")]
		public abstract string? keybinding{
			get;
			construct set;
		}
	}
	
	
	/**
	 * Actions which can be attached to user interface
	 * (a menu or a tool bar item, systray, jumplists).
	 */
	public class NormalAction: Gtk.Action, Action{
		[Description(nick = "Local keybinding", blurb = "Keyboard shortcut accessible inside window.")]
		public string? keybinding{
			get;
			construct set;
		}
		
		/**
		 * Creates new NormalAction object.
		 * 
		 * @param name			Unique name of an action
		 * @param label			Label visible to user
		 * @param tooltip		Tooltip text
		 * @param stock_id		Icon id
		 * @param keybinding	Local keyboard shortcut
		 */
		public NormalAction(string name, string? stock_id, string? label,  
			string? keybinding, string? tooltip, SimpleActionCallback? callback=null
		){
			Object(
				name: name,
				label: label,
				tooltip: tooltip,
				stock_id: stock_id,
				keybinding: keybinding
			);
			if(stock_id != null && Gtk.IconTheme.get_default().has_icon(stock_id)){
				icon_name = stock_id;
			}
			
			if(callback != null){
				activate.connect(() => {
					callback();
				});
			}
		}
	}
	
	
	
	/**
	 * Action with two states.
	 */
	public class ToggleAction: Gtk.ToggleAction, Action{
		[Description(nick = "Local keybinding", blurb = "Keyboard shortcut accessible inside window.")]
		public string? keybinding{
			get;
			construct set;
		}
		
		/**
		 * Creates new ToggleAction object.
		 * 
		 * @param name			Unique name of an action
		 * @param label			Label visible to user
		 * @param tooltip		Tooltip text
		 * @param stock_id		Icon id
		 * @param keybinding	Local keyboard shortcut
		 * @param active		Whether this option is toggled on
		 */
		public ToggleAction(string name, string? stock_id, string? label,  
			string? keybinding, string? tooltip, bool active=false
		){
			Object(
				name: name,
				label: label,
				tooltip: tooltip,
				stock_id: stock_id,
				keybinding: keybinding,
				active: active
			);
			if(stock_id != null && Gtk.IconTheme.get_default().has_icon(stock_id)){
				icon_name = stock_id;
			}
		}
	}
	
	/**
	 * Action with only one active item in a group.
	 */
	public class RadioAction: Gtk.RadioAction, Action{
		[Description(nick = "Local keybinding", blurb = "Keyboard shortcut accessible inside window.")]
		public string? keybinding{
			get;
			construct set;
		}
		
		/**
		 * Creates new RadioAction object.
		 * 
		 * @param name			Unique name of an action
		 * @param label			Label visible to user
		 * @param tooltip		Tooltip text
		 * @param stock_id		Icon id
		 * @param keybinding	Local keyboard shortcut
		 * @param value			Value of this option
		 */
		public RadioAction(string name, string? stock_id, string? label,  
			string? keybinding, string? tooltip, int value
		){
			Object(
				name: name,
				label: label,
				tooltip: tooltip,
				stock_id: stock_id,
				keybinding: keybinding,
				value: value
			);
			if(stock_id != null && Gtk.IconTheme.get_default().has_icon(stock_id)){
				icon_name = stock_id;
			}
		}
	}
	
	public class ActionCheckbox: Gtk.CheckButton{
		private Action action;
		
		public ActionCheckbox(Action action){
			Object(label: action.label.replace("_", ""));
			this.action = action;
		}
		
		public void run(){
			if(active && action.sensitive){
				action.activate();
			}
		}
	}
	
	public class ActionButton: Gtk.Button{
		private Action action;
		
		public ActionButton(Action action){
			Object(label: action.label.replace("_", ""));
			this.action = action;
			clicked.connect(on_clicked);
		}
		
		private void on_clicked(){
			if(action.sensitive){
				action.activate();
			}
		}
	}
	
	public class ActionImageButton: Widgets.ImageButton{
		private Action action;
		
		public ActionImageButton(Action action, Gtk.IconSize icon_size=Gtk.IconSize.BUTTON){
			base(action.icon_name ?? action.stock_id, action.label.replace("_", ""), icon_size);
			this.action = action;
			clicked.connect(on_clicked);
		}
		
		private void on_clicked(){
			if(action.sensitive){
				action.activate();
			}
		}
	}
	
	public class ActionGroups: GLib.Object
	{
		private HashTable<string, ActionGroup?> table;
		
		public ActionGroups()
		{
			table = new HashTable<string, ActionGroup?>(str_hash, str_equal);
		}
		
		public Action[] get_actions_of_group(string group)
		{
			var actions = table[group];
			if (actions == null)
				return {};
			return actions.get_actions();
		}
		
		public void add_actions_to_group(Action[] actions, string group)
		{
			foreach (var action in actions)
				add_action_to_group(action, group);
		}
		
		public virtual signal void add_action_to_group(Action action, string group)
		{
			var actions = table.lookup(group);
			if (actions == null)
			{
				actions = new ActionGroup();
				table.insert(group, actions);
			}
			actions.add(action);
			
		}
		
		public virtual signal void remove_action_from_group(Action action, string group)
		{
			var actions = table.lookup(group);
			if (actions != null)
				actions.remove(action);
		}
	}
	
	private class ActionGroup
	{
		private SList<Action> actions = null;
		
		public ActionGroup()
		{
		}
		
		public ActionGroup.with_actions(Action[] actions)
		{
			foreach (var action in actions)
				this.actions.prepend(action);
			this.actions.reverse();
		}
		
		public void add(Action action)
		{
			actions.append(action);
		}
		
		public void remove(Action action)
		{
			actions.remove(action);
		}
		
		public Action[] get_actions()
		{
			Action[] result = new Action[actions.length()];
			var i = 0;
			foreach (var action in actions)
				result[i++] = action;
			return result;
		}
	}
} // namespace Diorite
