/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;
namespace Diorite{
	public errordomain AppError{
		NOT_RUNNING,
		ALREADY_RUNNING,
		UNABLE_TO_TERMINATE,
		UNABLE_TO_ACTIVATE
	}
	
	private const int XFCE_SESSION_END = 4;
	private const string XFCE_SESSION_SERVICE_NAME = "org.xfce.SessionManager";
	private const string XFCE_SESSION_SERVICE_OBJECT = "/org/xfce/SessionManager";
	
	private const string GNOME_SESSION_SERVICE_NAME = "org.gnome.SessionManager";
	private const string GNOME_SESSION_SERVICE_OBJECT = "/org/gnome/SessionManager";

	public abstract class Application : Gtk.Application{
		private static Application? instance;
		public string desktop_entry{ get; protected set;}
		public string app_name{ get; protected set;}
		public string path_name{ get; protected set;}
		public string display_name{ get; protected set;}
		public string icon{ get; protected set; default = "";}
		public string version{ get; protected set; default = "";}
		private XfceSessionManager? xfce_session = null;
		private GnomeSessionManager? gnome_session = null;
		private GnomeSessionClientPrivate? gnome_session_client = null;
		
		public Application(string name, string desktop_entry, string path_name,
		                   string? startup_id=null){
			Object(application_id: name);
			this.display_name = name;
			this.app_name = name;
			this.desktop_entry = desktop_entry;
			this.path_name = path_name;
		}
		
		public signal void show_user_info(string title, string text);
		
		public override void startup(){
			hold();
			/* Set program name */
			System.prctl(15, app_name, 0, 0, 0);
			GLib.Environment.set_application_name(app_name);
			Gdk.set_program_class(app_name); // must be set after Gtk.init()!
			
			instance = this;
			Posix.signal (Posix.SIGINT, terminate_handler);
			Posix.signal (Posix.SIGTERM, terminate_handler);
			Posix.signal (Posix.SIGHUP, terminate_handler);
			
			Bus.watch_name(BusType.SESSION, XFCE_SESSION_SERVICE_NAME,
				BusNameWatcherFlags.NONE,
				xfce_session_appeared, xfce_session_vanished);
			
			Bus.watch_name(BusType.SESSION, GNOME_SESSION_SERVICE_NAME,
				BusNameWatcherFlags.NONE,
				gnome_session_appeared, gnome_session_vanished);
			base.startup();
		}
		
		/**
		 * Quits application
		 * 
		 * Notice: This method differs from method GLib.Application.quit() (proviced
		 * by GIO since version 2.32). If you called method hold(), you must call
		 * also method release. In C, make sure you call diorite_application_quit()
		 * instead of g_application_quit().
		 */
		public new void quit(){
			// TODO: transfer to GLib.Application.quit(), when required dependency
			// is gio >= 2.32
			if(!is_remote){
				release();
			}
			else{
				lib_critical("Remote quit not supported");
			}
		}
		
		private static void terminate_handler(int sig_num){
			lib_debug("Caught signal %d, exiting ...", sig_num);
			if(instance == null){
				lib_error("No instance to terminate");
			}
			instance.quit();
		}
		
		private void xfce_session_appeared(DBusConnection conn, string name, string owner){
			lib_debug("XFCE session appeared: %s, %s", name, owner);
			try{
				this.xfce_session = Bus.get_proxy_sync(BusType.SESSION,
					XFCE_SESSION_SERVICE_NAME,
					XFCE_SESSION_SERVICE_OBJECT);
				xfce_session.state_changed.connect(xfce_session_state_changed);
			}
			catch(IOError e){
				lib_warning("Unable to get proxy for Xfce session: %s", e.message);
				this.xfce_session = null;
			}
		}
		
		/**
		 * Removes proxy object
		 */
		private void xfce_session_vanished(DBusConnection conn, string name){
			lib_debug("XFCE session vanished: %s", name);
			if(xfce_session == null) return;
			xfce_session.state_changed.disconnect(xfce_session_state_changed);
			xfce_session = null;
		}
		
		private void xfce_session_state_changed(uint32 old_value, uint32 new_value){
			if(new_value == XFCE_SESSION_END){
				lib_debug("XFCE session end");
				quit();
			}
		}
		
		private void gnome_session_appeared(DBusConnection conn, string name, string owner){
			lib_debug("GNOME session appeared: %s, %s", name, owner);
			try{
				gnome_session = Bus.get_proxy_sync(BusType.SESSION,
					GNOME_SESSION_SERVICE_NAME,
					GNOME_SESSION_SERVICE_OBJECT);
				
				try{
					var client_object = gnome_session.register_client(app_name, "");
					gnome_session_client = Bus.get_proxy_sync(BusType.SESSION,
						GNOME_SESSION_SERVICE_NAME,
						client_object);
					gnome_session_client.stop.connect(gnome_session_client_stop);
					gnome_session_client.end_session.connect(gnome_session_client_end);
					gnome_session_client.query_end_session.connect(gnome_session_client_query);
				}
				catch(IOError e){
					lib_warning("Unable to get proxy for GNOME session client: %s", e.message);
					gnome_session = null;
				}
			}
			catch(IOError e){
				lib_warning("Unable to get proxy for GNOME session: %s", e.message);
				gnome_session = null;
			}
		}
		
		/**
		 * Removes proxy object
		 */
		private void gnome_session_vanished(DBusConnection conn, string name){
			lib_debug("GNOME session vanished: %s", name);
			if(gnome_session == null) return;
			gnome_session = null;
			
			if(gnome_session_client == null) return;
			gnome_session_client.stop.disconnect(gnome_session_client_stop);
			gnome_session_client.end_session.disconnect(gnome_session_client_end);
			gnome_session_client.query_end_session.disconnect(gnome_session_client_end);
			gnome_session_client = null;
			
		}
		
		private void gnome_session_client_stop(){
			lib_debug("GNOME Session client signal: %s", "stop");
			quit();
		}
		
		private void gnome_session_client_query(uint flags){
			lib_debug(@"GNOME Session client signal: %s, $flags", "query-end-session");
			if(gnome_session_client == null) return;
			try{
				gnome_session_client.end_session_response(true, "");
			}
			catch(IOError e){
				lib_warning("Unable to send end session response for GNOME session client: %s", e.message);
			}
		}
		
		private void gnome_session_client_end(uint flags){
			lib_debug(@"GNOME Session client signal: %s, $flags", "end-session");
			if(gnome_session_client == null) return;
			try{
				gnome_session_client.end_session_response(true, "");
			}
			catch(IOError e){
				lib_warning("Unable to send end session response for GNOME session client: %s", e.message);
			}
		}
	}
}


[DBus(name = "org.xfce.Session.Manager")]
private interface XfceSessionManager : Object{
	public signal void state_changed(uint32 old_value, uint32 new_value);
}

// http://people.gnome.org/~mccann/gnome-session/docs/gnome-session.html#org.gnome.SessionManager
[DBus(name = "org.gnome.SessionManager")]
private interface GnomeSessionManager : Object{
	public signal void session_over();
	public abstract string register_client(string app_id, string startup_id) throws IOError;
//~ 	public abstract string unregister_client(string client_id) throws IOError;
}

// http://people.gnome.org/~mccann/gnome-session/docs/gnome-session.html#org.gnome.SessionManager.ClientPrivate
[DBus(name = "org.gnome.SessionManager.ClientPrivate")]
private interface GnomeSessionClientPrivate : Object{
	public signal void stop();
	public signal void end_session(uint flags);
	public signal void query_end_session(uint flags);
	public signal void cancel_end_session();
	public abstract void end_session_response(bool is_ok, string reason) throws IOError;
}
