/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;
namespace Diorite.Test{
	// Used to terminate when serious fail occurred
	[CCode (cname="exit")]
	extern void exit(int exit_code);
	// List of messages from failed assertions
	Gee.List<string> fails;
	int tasks_pos = 0;
	Task? current_task;
	// Map used to obtain task by its id for a dependency check
	Gee.Map<string, Task> tasks_map;

	// Counter of passed assertions
	uint assertions_passed;
	// Counter of passed tests
	uint tests_passed;
	// Counter of failed tests
	uint tests_failed;
	// Counter of skipped tests
	uint tests_skipped;
	//Currently running task
	
	//Array of registered test cases
	TestCase[] test_cases;
	// Position in test cases array
	int test_cases_pos = 0;
	TestCase? current_test_case = null;
	
	bool initialized = false;
	MainLoop? loop = null;
	Callback? loop_run = null;
	Callback? loop_quit = null;
	
	// State of task
	public enum State{
		READY, // task is ready to run
		PASSED, // task passed
		FAILED, // task failed
		SKIPPED; // task skipped
	}
	
	public errordomain TestError{
		DEPENDENCY_UNSATISFIED,
		ASSERTION_FAILED
	}
	
	/**
	 * Signature of test function
	 */
	public delegate void TestFunc() throws TestError;
	public delegate void Callback();
	
	/**
	 * Initializes unit test framework
	 * @param args Command line arguments
	 * @param loop_run function to start main loop (e.g. Gtk.main)
	 * @param loop_quit function to quit main loop (e.g. Gtk.main_quit)
	 */
	public static void init(string[] args,
		owned Callback? loop_run=null, owned Callback? loop_quit=null){
		if(loop_run == null && loop_quit != null
		|| loop_run != null && loop_quit == null){
			lib_error("You must provide both loop_run and loop_quit or none of them.");
		}
		Test.loop_run = (owned) loop_run;
		Test.loop_quit = (owned) loop_quit;
		fails = new Gee.ArrayList<string>();
		tasks_map = new Gee.HashMap<string, Task>();
		test_cases = {};
		current_test_case = null;
		tests_passed = 0;
		assertions_passed = 0;
		tests_failed = 0;
		tests_skipped = 0;
		initialized = true;
		
		
		/* Set up data directories */
		File? dir = File.new_for_path("./");
		if(dir.get_basename() == "build") dir = dir.get_parent();
		else if(!dir.get_child("build").query_exists()) dir = null;
		
		if(dir != null){
			string dirs = dir.get_child("tests/dirs/data_first").get_path()
			+ ":" + dir.get_child("data").get_path()
			+ ":" + dir.get_child("build/data").get_path()
			+ ":" + dir.get_child("tests/dirs/data_last").get_path();
			
			lib_debug("XDG_DATA_DIRS = %s", dirs);
			Environment.set_variable("XDG_DATA_DIRS", dirs, true);
			var user_data = dir.get_child("tests/dirs/data_user").get_path();
			
			lib_debug("XDG_DATA_HOME = %s", user_data);
			Environment.set_variable("XDG_DATA_HOME", user_data, true);
		}
		else{
			lib_error("Unable to find source directory. Current location: %s",
				dir.get_path());
		}
	}
	
	/**
	 * Adds new test case
	 * @param test_case Glib type of the test case, must be a subclass of Yauty.TestCase
	 */
	public static void add(GLib.Type test_case_type){
		if(!initialized){
			lib_error("Test framework not initialized. Use Test.init");
		}
		var test_case = GLib.Object.new(test_case_type) as TestCase;
		if(test_case == null){
			// Oh, somebody tried to trick me!
			lib_error("%s is not subclass of TestCase.\n", test_case_type.name());
		}
		test_cases += test_case;
		foreach(var task in test_case.tasks){
			tasks_map[task.id] = task;
		}
	}
	
	/**
	 * Runs test cases.
	 */
	public static int run(){
		if(!initialized){
			error("Test framework not initialized. Use Test.init");
		}
		
		GLib.Test.timer_start();
		
		if(loop_run == null){
			loop = new MainLoop();
		}
		
		next_test_case();
		
		if(loop_run == null){
			loop.run();
		}
		else{
			loop_run();
		}
		
		var time = GLib.Test.timer_elapsed();
		
		/* Print error details */
		if(tests_failed > 0){
			int i = 1;
			stderr.printf(@"\n============ ERRORS: $tests_failed ============\n");
			foreach(var error in fails){
				stderr.printf(@"\n===== $i) $error\n\n");
				i++;
			}
		}
		
		/* Print result */
		string result;
		int retcode;
		if(tests_failed == 0){
			result = Colors.BLD_GREEN + "OK" + Colors.RESET;
			retcode = 0;
		}
		else{
			result = Colors.BLD_RED + "FAIL" + Colors.RESET;
			retcode = 1;
		}
		stderr.printf(@"\n============ RESULT: $result ============\nPassed tests: $tests_passed\nFailed tests: $tests_failed\nSkipped tests: $tests_skipped\nPassed assertions: $assertions_passed\nTime elapsed: $time\n");
		return retcode;
	}
	
	private void next_test_case(){
		// Call cleaning-up method for previous test case
		if(current_test_case != null){
			current_test_case.tear_down_after();
		}
		
		// Return if all tests are finished
		if(test_cases_pos == test_cases.length){
			if(loop_quit == null){
				loop.quit();
			}
			else{
				loop_quit();
			}
			return;
		}
		
		// Start new test case
		current_test_case = test_cases[test_cases_pos++];
		current_test_case.set_up_before();
		tasks_pos = 0;
		Idle.add(next_task);
	}
	
	bool next_task(){
//~ 		stderr.printf("tast pos: %d/%d\n", task_pos, tasks.size);
		if(current_task != null){
			current_task.test.tear_down();
		}
		
		if(tasks_pos == current_test_case.tasks.size){
			next_test_case();
			return false;
		}
		
		var task = current_test_case.tasks[tasks_pos++];
		try{
			if(task.dependencies.length != 0){
				foreach(var dependency in task.dependencies){
					Task? dep = tasks_map[dependency];
					
					if(dep == null){
						throw new TestError.DEPENDENCY_UNSATISFIED(
							@"Dependency '$dependency' not found");
					}
					
					switch(dep.status){
					case State.SKIPPED:
						throw new TestError.DEPENDENCY_UNSATISFIED(
							@"Dependency '$dependency' was skipped");
					case State.READY:
						throw new TestError.DEPENDENCY_UNSATISFIED(
							@"Dependency '$dependency' has not been executed yet");
					case State.FAILED:
						throw new TestError.DEPENDENCY_UNSATISFIED(
							@"Dependency '$dependency' failed");
					}
				}
			}
			
			/* Set up task */
			task.test.set_up();
			current_task = task;
			stderr.printf("\n%s=====%s /%s/%s ... running\n", Colors.BLD_WHITE,
					Colors.RESET, task.test.name, task.name);
			
			task.run();
			
			stderr.printf("%s=====%s /%s/%s ... %sOK%s\n", Colors.BLD_WHITE,
					Colors.RESET, task.test.name, task.name,
				Colors.BLD_GREEN, Colors.RESET);
			task.status = State.PASSED;
			tests_passed++;
		}
		catch(TestError e){
			if(e is TestError.ASSERTION_FAILED || e is TestError.DEPENDENCY_UNSATISFIED){
				string e_name = null;
				if(e is TestError.DEPENDENCY_UNSATISFIED){
					stderr.putc('\n');
					e_name = "Unsatisfied dependency";
				}
				else{
					e_name = "Assertion error";
				}
				fails.add("/%s/%s \t\n%s: %s".printf(
					task.test.name, task.name, e_name, e.message));
				stderr.printf("%s=====%s /%s/%s ... %sFAIL%s\n", Colors.BLD_WHITE,
					Colors.RESET, task.test.name, task.name,
					Colors.BLD_RED, Colors.RESET);
				task.status = State.FAILED;
				tests_failed++;
			}
			else{
				stderr.printf("\n\n Unexpected error: %s\n", e.message);
				exit(1);
			}
		}
		
		Idle.add(next_task);
		return false;
	}
	
	/**
	 * This namespace contains base assertion primitives.
	 */
	namespace Assert{
		/**
		 * Asserts that condition is true
		 * 
		 * @param condition	condition to test
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void is_true(bool condition, string? message = null) throws TestError{
			if(!condition){
				fail(message ?? "Assertion 'condition is true' failed.");
			}
			assertions_passed++;
		}
		
		/**
		 * Asserts that condition is false
		 * 
		 * @param condition	condition to test
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void is_false(bool condition, string? message = null) throws TestError{
			is_true(!condition, message ?? "Assertion 'condition is false' failed.");
		}
		
		/**
		 * Asserts that the code has not been reached
		 * 
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void not_reached(string? message = null) throws TestError{
			fail(message ?? "Assertion 'not reached' failed.");
		}
		
		/**
		 * Asserts that object is null
		 * 
		 * @param object	object to test
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void is_null(GLib.Object? object, string? message = null) throws TestError{
			is_true(object == null, message ?? "Assertion 'object is null' failed.");
		}
		
		/**
		 * Asserts that object is not null
		 * 
		 * @param object	object to test
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void not_null(GLib.Object? object, string? message = null) throws TestError{
			is_true(object != null, message ?? "Assertion 'object is not null' failed.");
		}
		
		/**
		 * Asserts that given comparison of two string values is true
		 * 
		 * @param s1		expected value
		 * @param op		operator (==, !=, <, <=, <, >=, =~, ~=, ~=~)
		 * @param s2		actual value
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void cmp_str(string? s1, string op, string? s2) throws TestError{
			bool result;
			var _s1 = s1 == null ? "(null)" : "'" + s1 + "'";
			var _s2 = s2 == null ? "(null)" : "'" + s2 + "'";
			switch(op){
			case "==": result = s1 == s2; break;
			case "!=": result = s1 != s2; break;
			case "<": result = s1 < s2; break;
			case "<=": result = s1 <= s2; break;
			case ">": result = s1 > s2; break;
			case ">=": result = s1 >= s2; break;
			case "=~": result = s1 != null && s2 != null && s1.has_prefix(s2); break;
			case "~=": result = s1 != null && s2 != null && s1.has_suffix(s2); break;
			case "~=~": result = s1 != null && s2 != null && (s2 in s1); break;
			default:
				fail(@"Unknown assertion for cmp_str: $_s1 $op $_s2.");
				return;
			}
			
			is_true(result, @"Assertion $_s1 $op $_s2 failed.");
		}
		
		/**
		 * Asserts that given comparison of two int values is true
		 * 
		 * @param i1		expected value
		 * @param op		operator (==, !=, <, <=, <, >=)
		 * @param i2		actual value
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void cmp_int(int i1, string op, int i2) throws TestError{
			bool result;
			switch(op){
			case "==": result = i1 == i2; break;
			case "!=": result = i1 != i2; break;
			case "<": result = i1 < i2; break;
			case "<=": result = i1 <= i2; break;
			case ">": result = i1 > i2; break;
			case ">=": result = i1 >= i2; break;
			default: 
				fail(@"Unknown assertion for cmp_int: '$i1 $op $i2'.");
				return;
			}
			
			is_true(result, @"Assertion '$i1 $op $i2' failed.");
		}
		
		/**
		 * Asserts that given comparison of two uint values is true
		 * 
		 * @param i1		expected value
		 * @param op		operator (==, !=, <, <=, <, >=)
		 * @param i2		actual value
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void cmp_uint(uint i1, string op, uint i2) throws TestError{
			bool result;
			switch(op){
			case "==": result = i1 == i2; break;
			case "!=": result = i1 != i2; break;
			case "<": result = i1 < i2; break;
			case "<=": result = i1 <= i2; break;
			case ">": result = i1 > i2; break;
			case ">=": result = i1 >= i2; break;
			default: 
				fail(@"Unknown assertion for cmp_uint: '$i1 $op $i2'.");
				return;
			}
			
			is_true(result, @"Assertion '$i1 $op $i2' failed.");
		}
		
		/**
		 * Asserts that given comparison of two double values is true
		 * 
		 * @param i1		expected value
		 * @param op		operator (==, !=, <, <=, <, >=)
		 * @param i2		actual value
		 * @param message	optional failure message
		 * @throw TestError on assertion failure
		 */
		public void cmp_double(double i1, string op, double i2) throws TestError{
			bool result;
			switch(op){
			case "==": result = i1 == i2; break;
			case "!=": result = i1 != i2; break;
			case "<": result = i1 < i2; break;
			case "<=": result = i1 <= i2; break;
			case ">": result = i1 > i2; break;
			case ">=": result = i1 >= i2; break;
			default: 
				fail(@"Unknown assertion for cmp_uint: '$i1 $op $i2'.");
				return;
			}
			
			is_true(result, @"Assertion '$i1 $op $i2' failed.");
		}
		
		/**
		 * Fails a test with a failure message.
		 * 
		 * @param message	failure message
		 * @throw TestError with given failure message
		 */
		public void fail(string message) throws TestError{
			throw new TestError.ASSERTION_FAILED(message);
		}
		
	}

	public class AsyncContext: GLib.Object{
			private bool event_done = false;
			private MainLoop? loop = null;
			private TestError? error = null;
			public ulong handler_id = 0;
			public uint event_timeout = 0;
			
			public void interrupt(){
				fail(new TestError.ASSERTION_FAILED("Test was interrupted"));
			}
			
			public void fail(TestError error){
				this.error = error;
				done();
			}
			
			public void done(){
				if(!event_done){
					event_done = true;
					if(loop != null){
						loop.quit();
						loop = null;
					}
				}
			}
			
			public void assert_done(string? message = null) throws TestError{
				if(event_done){
					event_done = false;
					check_error();
				}
				else{
					check_error();
					Assert.fail(message ?? "Assertion 'event is done' failed.");
				}
			}
			
			private void check_error() throws TestError{
				var error = this.error;
				if(error != null){
					this.error = null;
					throw error;
				}
			}
			
			public void assert_timeout(uint seconds = 1, string? message = null) throws TestError{
				if(event_done){
					event_done = false;
					check_error();
					return;
				}
				
				loop = new MainLoop();
				event_timeout = GLib.Timeout.add_seconds(seconds, () => {
					loop.quit();
					return false;
				});
				
				loop.run();
				Source.remove(event_timeout);
				assert_done(message ?? @"Loop tiemout $seconds second(s) reached.");
				return;
			}
		}
		
	
	
	/**
	 * Base class for all test cases. Test case can contain one or more test methods (also
	 * called tasks). Test methods should be prefixed by "test_", e.g. "test_parse_data()".
	 * Each test method can contain one or more assertions provided by functions
	 * in Diorite.Test.Assert namespace. The test method have to be registered
	 * in the construct block by add_task() method.
	 * 
	 * 
	 * {{{
	 *  using Diorite.Test;
	 *  
	 *  public class MyCaseOneTest: Diorite.Test.TestCase{
	 *  	construct{
	 * 			// Task without dependency
	 *  		add_task("foo", test_foo);
	 *  		// Task dependent on task "foo" (relative)
	 *  		add_task("bar", test_bar, {"foo"});
	 * 			// Task dependent on task "foo" (absolute)
	 * 			add_task("bar", test_bar, {"/MyCaseOneTest/foo"});
	 * 			// Task dependent on task "bar" from another test case
	 * 			add_task("bar", test_bar, {"/MyCaseTwoTest/bar"});
	 * 		}
	 * 		
	 * 		public void test_foo() throws TestError{
	 * 			Assert.cmp_str("foo", "!=", "bar");
	 * 			Assert.cmp_int(1, "<=", 3);
	 * 		}
	 * 		
	 * 		public void test_bar() throws TestError{
	 * 			Assert.str_cmp("foo", ">", "bar");
	 * 		}
	 *  }
	 * }}}
	 * 
	 * You may also override method set_up(), set_up_before(), tear_down()
	 * and tear_down_after() to prepare data and perform clean-up.
	 */
	public abstract class TestCase: Object{
		public string name{
			get{
				return this.get_type().name();
			}
		}
		
		public Gee.List<Task> tasks = new Gee.ArrayList<Task>();
		
		/**
		 * Add task to this test case
		 * @param name			Name of the task (should be similar to method name)
		 * @param func			Test function (method of the class)
		 * @param dependencies	Identifiers of the test tasks this function
		 * 		is dependent on. It can be relative to this test case ("test_method_1")
		 * 		or absolute ("/FooTestCase/test_method_1").
		 */
		protected void add_task(string name, TestFunc func, string[]? dependencies=null){
			if(dependencies != null){
				for(int i = 0; i < dependencies.length; i++){
					string dep = dependencies[i];
					if(dep.length == 0 || dep.length == 1 && dep[0] == '/'){
						stderr.printf("ERR: Invalid empty dependency #%d for %s.%s\n",
							i, this.name, name);
						exit(1);
					}
					if(dep[0] != '/'){
						dependencies[i] = "/%s/%s".printf(this.name, dep);
					}
				}
			}
			string[] empty_dependencies = {};
			var task = new Task(this, name, func, dependencies ?? empty_dependencies); 
			tasks.add(task);
		}
		
		/**
		 * This function is lauched before the first task of this test case is launched.
		 * It's generally used to prepare data shared by all tasks.
		 */
		public virtual void set_up_before(){
		}
		
		/**
		 * This function is called right before each task is launched
		 * (including the first task).
		 */
		public virtual void set_up(){
		}
		
		/**
		 * This function is called right after each task is finished
		 * (including the latest task).
		 */
		public virtual void tear_down(){
		}
		
		/**
		 * This function is called after the latest task is finished.
		 * It's generally used to clean-up the environment.
		 */
		public virtual void tear_down_after(){
		}
		
	}
		
	
	/**
	 * Class representing single test task
	 */
	public class Task{
		public string id{
			owned get{
				return "/%s/%s".printf(test.name, name);
			}
		}
		/** Test case task belongs to */
		public TestCase test;
		/** Name of the task */
		public string name;
		/** Test function */
		public unowned TestFunc run;
		/** State of the task */
		public State status = State.READY;
		/** Id of dependecy task */
		public string[] dependencies;
		
		/** 
		 * Creates new test task
		 * @param test			Test case task belongs to
		 * @param name			Name of the task
		 * @param func			Test function
		 * @param dependencies	Identifiers of dependencies
		 */
		public Task(TestCase test, string name, TestFunc func, string[] dependencies){
			this.name = name;
			this.test = test;
			this.run = func;
			this.dependencies = dependencies;
		}
	}
	
	namespace Colors{
		public const string RESET = "\x001b[0m";
		public const string GREEN = "\x001b[0;32m";
		public const string RED = "\x001b[0;31m";
		public const string YELLOW = "\x001b[0;33m";
		public const string BLD_WHITE = "\x001b[1;37m";
		public const string BLD_GREEN = "\x001b[1;32m";
		public const string BLD_RED = "\x001b[1;31m";
		public const string BLD_YELLOW = "\x001b[1;33m";
	}
}
