/*
 * Copyright 2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Diorite
{

/**
 * SimpleUIParser parses a limited subset of Gtk.UIManager XML format and creates
 * Gtk widgets.
 * 
 * Only elements <ui>, <menubar>, <menu>, <menuitem> and <separator> are
 * supported. <menubar>, <menu> and <menuitem> elements have to have
 * attribute action. Behavior with invalid input data is undefined.
 * 
 * Unlike Gtk.UIManager, SimpleUIParser don't support merging of multiple
 * UI files.
 */
public class SimpleUIParser: GLib.Object
{
	const MarkupParser parser = { start_tag, end_tag, null, null, null};
	private MarkupParseContext context;
	private Queue<Gtk.MenuItem> menu_chain = new Queue<Gtk.MenuItem>();
	private Gtk.MenuShell? root = null;
	private Gtk.MenuShell? menu = null;
	private Gtk.MenuItem menu_item = null;
	private Diorite.Actions actions = null;
	private HashTable<string, Gtk.Widget> widgets = new HashTable<string, Gtk.Widget>(str_hash, str_equal);
    
	/**
	 * Creates new SimpleUIParser
	 * 
	 * @param actions		actions to be used by parser for menu items
	 */
	public SimpleUIParser(Diorite.Actions actions)
	{
		this.actions = actions;
	}
	
	/**
	 * Loads UI tree from file.
	 * 
	 * @param ui_file		UI tree file to load
	 * @throws MarkupError on failure
	 */
	public void load_from_file(File ui_file) throws MarkupError
	{
		string ui_xml;
		try
		{
			ui_xml = Diorite.System.read_file(ui_file);
		}
		catch (Error e)
		{
			throw new MarkupError.INVALID_CONTENT("Unable to read file %s.", ui_file.get_path());
		}
		load(ui_xml);
	}
	
	/**
	 * Loads UI tree from string.
	 * 
	 * @param ui_file		UI tree string to load
	 * @throws MarkupError on failure
	 */
	public void load(string ui_xml) throws MarkupError
	{
		menu = null;
		menu_item = null;
		root = null;
		menu_chain.clear();
		widgets.remove_all();
		
		context = new MarkupParseContext(parser, 0, this, destroy);
		context.parse(ui_xml, -1);
	}
	
	private void start_tag(MarkupParseContext context, string name,
	string[] attr_names, string[] attr_values) throws MarkupError
	{
//~ 		debug("Start tag: %s", name);
		int i;
		switch (name)
		{
		case "menubar":
			for(i = 0; i < attr_names.length; i++){
				if(attr_names[i] == "action"){
					break;
				}
			}
			if(i >= attr_values.length)
			{
				throw new MarkupError.MISSING_ATTRIBUTE("Missing attribute 'action' for element 'menubar'.");
			}
			
			root = new Gtk.Menu();
			widgets.insert(attr_values[i], root);
			menu = root;
			menu_item = null;
			break;
		case "menu":
			if (menu == null)
			{
				throw new MarkupError.INVALID_CONTENT("Menu element requires parent element.");
			}
			
			for(i = 0; i < attr_names.length; i++){
				if(attr_names[i] == "action"){
					break;
				}
			}
			if(i >= attr_values.length)
			{
				throw new  MarkupError.MISSING_ATTRIBUTE("Missing attribute 'action' for element 'menu'.");
			}
			
			var action = actions.get_action(attr_values[i]);
			return_if_fail(action != null);
			var new_menu_item = action.create_menu_item() as Gtk.MenuItem;
			return_if_fail(new_menu_item != null);
			
			menu.append(new_menu_item);
			new_menu_item.show();
			new_menu_item.submenu = new Gtk.Menu();
			
			if (menu_item != null)
			{
				menu_chain.push_tail(menu_item);
			}
			menu_item = new_menu_item;
			menu = menu_item.submenu;
			
			break;
		case "menuitem":
			if (menu == null)
			{
				throw new MarkupError.INVALID_CONTENT("Menuitem element requires parent element.");
			}
			for(i = 0; i < attr_names.length; i++){
				if(attr_names[i] == "action"){
					break;
				}
			}
			if(i >= attr_values.length)
			{
				throw new  MarkupError.MISSING_ATTRIBUTE("Missing attribute 'action' for element 'menuitem'.");
			}
			var action = actions.get_action(attr_values[i]);
			return_if_fail(action != null);
			var new_menu_item = action.create_menu_item() as Gtk.MenuItem;
			return_if_fail(new_menu_item != null);
			menu.append(new_menu_item);
			new_menu_item.show();
			break;
		case "separator":
			var separator = new Gtk.SeparatorMenuItem();
			menu.append(separator);
			separator.show();
			break;
		default:
			debug("SimpleUIParser: ignored %s", name);
			break;
		}
		
	}
	
	private void end_tag(MarkupParseContext context, string name)
	throws MarkupError
	{
//~ 		debug("End tag: %s", name);
		if(name == "menu"){
			if (menu_chain.is_empty())
			{
				menu_item = null;
				menu = root;
			}
			else{
				menu_item = menu_chain.pop_tail();
				menu = menu_item.submenu;
			}
		}
		else if(name == "menubar")
		{
			menu_item = null;
			menu = null;
			root = null;
			menu_chain.clear();
		}
	}
	
	private void destroy(){
		debug("Destroy");
		menu = null;
		menu_item = null;
		menu_chain.clear();
	}
	
	/**
	 * Returns a top level widget
	 * 
	 * @param name		widget name from param "action"
	 * @return			a widget or null on failure
	 */
	public unowned Gtk.Widget? get_widget(string name){
		return widgets.lookup(name);
	}
}

} // namespace Diorite

