/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola{

/**
 * FrameBridge provides bridge between Vala code and JavaScript environment for a given frame.
 * 
 * It provides methods to execute scripts in that frame with {@link object_this} used in context
 * of 'this'.
 */
public class FrameBridge: GLib.Object
{
	private unowned JS.Object? _object_this = null;
	private unowned JS.Context? _context = null;
	
	[Description(nick = "Frame id", blurb = "Identifier of this frame.")]
	public string frame_id{get; private set;}
	
	[Description(nick = "JavaScript API", blurb = "JavaScript API associated with this bridge.")]
	public JSApi js_api{get; private set;}
	
	[Description(nick = "Web frame", blurb = "Web frame associated with this bridge.")]
	public weak WebKit.WebFrame frame{get; private set;}
	
	/**
	 * Emitted when page has been loaded.
	 */ 
	public signal void load_finished();
	
	/**
	 * Creates new FrameBridge.
	 * 
	 * @param frame_id       custom identifier of the frame
	 * @param frame          frame to be used for bridge
	 * @param js_api         JavaScript API to be used for bridge
	 * @param object_this    object to be used in context of 'this'
	 */
	public FrameBridge(string frame_id, WebKit.WebFrame frame, JSApi js_api)
	{
		this.frame_id = frame_id;
		this.frame = frame;
		this.js_api = js_api;
		set_object(get_real_context(), null);
		frame.notify["load-status"].connect(on_load_status_changed);
		#if DEBUG_MEMORY
		debug("new FrameBridge, js_api refs: %u", js_api.ref_count);
		#endif
	}
	
	#if DEBUG_MEMORY
	~FrameBridge()
	{
		debug("~FrameBridge, js_api refs: %u", js_api.ref_count);
	}
	#endif
	
	public void clear_object()
	{
		set_object(null, null);
	}
	
	
	public bool ref_object(out unowned JS.Context? ctx, out unowned JS.Object? object)
	{
		lock (_context)
		{
			ctx = null;
			object = null;
			if (_object_this == null)
				return false;
			ctx = get_context();
			if (ctx == null)
				return false;
			
			ctx.retain();
			_object_this.protect(ctx);
			object = _object_this;
			return true;
		}
	}
	
	public bool unref_object(JS.Context ctx, JS.Object object)
	{
		lock (_context)
		{
			if (ctx != get_context() || ctx == null || _object_this == null || _object_this != object)
				return false;
			
			object.unprotect(ctx);
			ctx.release();
			return true;
		}
	}
	
	public void set_object(JS.Context? ctx=null, JS.Object? object_this=null)
	{
		lock (_context)
		{
			/* Protect new context and object */
			if (ctx != null)
			{
				ctx.retain();
				if (object_this != null)
					object_this.protect(ctx);
			}
			
			get_context();
			
			/* Unprotect old context and object */
			if (_context != null)
			{
				if (_object_this != null)
					_object_this.unprotect(_context);
				_context.release();
			}
			
			_context = ctx;
			_object_this = object_this;
		}
	}
	
	private unowned JS.Context get_context()
	{
		return _context;
	}
	
	/**
	 * Returns real JavaScript context for the frame.
	 * 
	 * @return    JavaScript context
	 */
	public unowned JS.Context get_real_context()
	{
		unowned JS.Context ctx = JS.Context.for_frame(frame);
		return ctx;
	}
	
	/**
	 * Executes script from file.
	 * 
	 * The script will be executed in a context returned by {@link get_context().
	 * The "this" keyword will refer to {@link object_this} if provided.
	 * 
	 * @param file    script to execute
	 * @return        return value of the script
	 * @throw         JSError on failure
	 */
	public unowned Value execute_script_from_file(File file) throws JSError
	{
		string code;
		try
		{
			code = Diorite.System.read_file(file);
		}
		catch (Error e)
		{
			throw new JSError.READ_ERROR("Unable to read script %s: %s",
				file.get_path(), e.message);
		}
		return execute_script(code, file.get_uri(), 1);
	}
	
	/**
	 * Executes script.
	 * 
	 * The script will be executed in a context returned by {@link get_context()}.
	 * The "this" keyword will refer to {@link object_this} if provided.
	 * 
	 * @param script    script to execute
	 * @return          return value of the script
	 * @throw           JSError on failure
	 */
	public unowned Value execute_script(string script, string path = "about:blank", int line=1) throws JSError
	{
		lock (_context)
		{
			unowned JS.Context ctx = get_context(); 
			JS.Value exception = null;
			unowned Value value = ctx.evaluate_script(new JS.String(script), _object_this, new JS.String(path), line=0, out exception);
			if (exception != null)
				throw new JSError.EXCEPTION(JSTools.exception_to_string(ctx, exception));
			return value;
		}
	}
	
	private void on_load_status_changed(GLib.Object o, ParamSpec p)
	{
		var frame = o as WebKit.WebFrame;
		return_if_fail(frame != null);
		if (frame.load_status == WebKit.LoadStatus.FINISHED)
			load_finished();
	}
}

} // namespace Nuvola
