/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using WebKit;
using Soup;
using Gee;
using Diorite;
using Diorite.Logger;
using Diorite.Widgets;
using Diorite.Tools;

namespace Nuvola{

public const string STATE_PLAYING = "playing";
public const string STATE_PAUSED = "paused";
public const string STATE_NONE = "none";


namespace Actions{
	public const string PLAY = "play";
	public const string PAUSE = "pause";
	public const string TOGGLE_PLAY = "toggle-play";
	public const string PREV_SONG = "prev";
	public const string NEXT_SONG = "next";
	public const string THUMBS_UP = "thumbs-up";
	public const string THUMBS_DOWN = "thumbs-down";
	public const string FAVORITE = "favorite";
}

const string BASE_PLAYER_ACTIONS = "player";
const string EXTRA_PLAYER_ACTIONS = "player-extra";
const string PLAYER_ACTIONS_EXTRA[3] = {Actions.THUMBS_UP, Actions.THUMBS_DOWN, Actions.FAVORITE};
const string PLAYER_ACTIONS[3] = {Actions.PREV_SONG, Actions.TOGGLE_PLAY, Actions.NEXT_SONG};
const string PLAYER_ACTIONS_ALL[6] = {Actions.PREV_SONG, Actions.TOGGLE_PLAY, Actions.NEXT_SONG, Actions.THUMBS_UP, Actions.THUMBS_DOWN, Actions.FAVORITE};


/**
 * This interface provides control of playback.
 */
public abstract class Player: GLib.Object{
	/// Action label
	static string PLAY_LABEL = _("Play");
	/// Action label
	static string PAUSE_LABEL = _("Pause");
	static string PLAY_ICON = "media-playback-start";
	static string PAUSE_ICON = "media-playback-pause";
	
	protected weak Diorite.Actions? actions;
	protected weak Diorite.ActionGroups? action_groups;
	
	[Description(nick = "Path to album art", blurb = "Path to album art")]
	public string? album_art {get; protected set;} 
	
	[Description(nick = "Path to album art", blurb = "Path to album art")]
	public string? album {get; protected set;} 
	
	[Description(nick = "name of the current artist", blurb = "name of the current artist")]
	public string? artist {get; protected set;} 
	
	[Description(nick = "name of the current song", blurb = "name of the current song")]
	public string? song {get; protected set;}
	
	[Description(nick = "playback state", blurb = "playback state: 'playing', 'paused' or 'none'")]
	
	private string _playback_state = STATE_NONE;
	public string playback_state{
		get{
			return _playback_state;
		}
		protected set{
			_playback_state = value;
			if(actions != null){
				var action_toggle = actions.get_action(Actions.TOGGLE_PLAY);
				var action_play = actions.get_action(Actions.PLAY);
				var action_pause = actions.get_action(Actions.PAUSE);
				
				if (action_toggle != null && action_play != null && action_pause != null){
					switch(_playback_state){
		 			case STATE_PAUSED:
						action_toggle.stock_id = PLAY_ICON;
						action_toggle.icon_name = PLAY_ICON;
						action_toggle.label = PLAY_LABEL;
						action_toggle.sensitive = true;
						action_play.sensitive = true;
						action_pause.sensitive = false;
						break;
					case STATE_PLAYING:
						action_toggle.stock_id = PAUSE_ICON;
						action_toggle.icon_name = PAUSE_ICON;
						action_toggle.label = PAUSE_LABEL;
						action_toggle.sensitive = true;
						action_play.sensitive = false;
						action_pause.sensitive = true;
						break;
					default:
						action_toggle.stock_id = PLAY_ICON;
						action_toggle.icon_name = PLAY_ICON;
						action_toggle.label = PLAY_LABEL;
						action_toggle.sensitive = false;
						action_play.sensitive = false;
						action_pause.sensitive = false;
						break;
					}
				}
				else
				{
					if (action_toggle == null)
						warning("Missing action: %s", Actions.TOGGLE_PLAY);
					if (action_play == null)
						warning("Missing action: %s", Actions.PLAY);
					if (action_pause == null)
						warning("Missing action: %s", Actions.PAUSE);
				}
			}
		}
	}
	
	/**
	 * Creates new JSApiPlayer object.
	 * 
	 * @param actions          user interface actions
	 * @param action_groups    user interface action groups
	 */
	public Player(Diorite.Actions? actions, Diorite.ActionGroups? action_groups)
	{
		this.actions = actions;
		this.action_groups = action_groups;
		if (actions != null)
		{
			append_ui_actions();
		}
	}
	
	/**
	 * Skips to the next track in the tracklist.
	 * 
	 * If there is no next track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void next_song();
	
	/***
	 * Skips to the previous track in the tracklist.
	 * 
	 * If there is no previous track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void previous_song();
	
	/**
	 * Pauses playback.
	 * 
	 * If playback is already paused, this has no effect. Calling Play after this should
	 * cause playback to start again from the same position.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void pause();
	
	/**
	 * Pauses playback.
	 * If playback is already paused, resumes playback.
	 * If playback is stopped, starts playback.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void toggle_play();
	
	/**
	 * Stops playback.
	 * 
	 * If playback is already stopped, this has no effect. Calling Play after this should
	 * cause playback to start again from the beginning of the track. 
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void stop();
	
	/**
	 * Starts or resumes playback.
	 * 
	 * If already playing, this has no effect.
	 * If there is no track to play, this has no effect.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void play();
	
	/**
	 * Sets Thumbs Up (= awesome) to song
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void thumbs_up();
	
	/**
	 * Sets Thumbs Dowb (= horrible) to song
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void thumbs_down();
	
	/**
	 * Marks song as favorited
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public abstract void favorite();
	
	/**
	 * Creates UI actions and appends them to UIActions instance
	 */
	protected virtual void append_ui_actions(){
		/* Normal items. Name, stock id, label, accelerator, tooltip, callback */
		
		Diorite.Action[] base_actions =
		{
			/// Action label
			new NormalAction(Actions.PREV_SONG, "media-skip-backward", _("Previous song"), 
			/// Action tooltip
			"<alt>Left", _("Skip to previous song"), previous_song),
			/// Action tooltip
			new NormalAction(Actions.TOGGLE_PLAY, PLAY_ICON, PLAY_LABEL, "<ctrl>space", _("Toggle play/pause"), toggle_play),
			/// Action label
			new NormalAction(Actions.NEXT_SONG, "media-skip-forward", _("Next song"), "<alt>Right", 
			/// Action tooltip
			_("Skip to next song"), next_song)
		};
		actions.add_actions(base_actions);
		action_groups.add_actions_to_group(base_actions, BASE_PLAYER_ACTIONS);
		
		Diorite.Action[] extra_actions =
		{
			/// Short label for command 'thumb up' or 'like this song' (see Google Music buttons)
			new NormalAction(Actions.THUMBS_UP, null, _("Thumbs Up"), null, null, thumbs_up),
			/// Short label for command 'thumb down' or 'dislike this song' (see Google Music buttons)
			new NormalAction(Actions.THUMBS_DOWN, null, _("Thumbs Down"), null, null, thumbs_down),
			/// Short label for command 'add this song to favorites'
			new NormalAction(Actions.FAVORITE, null, _("Favorite"), null, null, favorite)
		};
		actions.add_actions(extra_actions);
		action_groups.add_actions_to_group(extra_actions, EXTRA_PLAYER_ACTIONS);
		
		actions.add_actions({
			new NormalAction(Actions.PLAY, PLAY_ICON, PLAY_LABEL, null, PLAY_LABEL, play),
			/// Tool bar item tooltip
			new NormalAction(Actions.PAUSE, PAUSE_ICON, PAUSE_LABEL, null, PAUSE_LABEL, pause)
		});
		
		actions.get_action(Actions.TOGGLE_PLAY).sensitive = false;
		actions.get_action(Actions.PREV_SONG).sensitive = false;
		actions.get_action(Actions.NEXT_SONG).sensitive = false;
		actions.get_action(Actions.PLAY).sensitive = false;
		actions.get_action(Actions.PAUSE).sensitive = false;
	}
	
	/**
	 * Emitted when song is changed.
	 * 
	 * @param song name of the current song
	 * @param artist name of the current artist
	 * @param album album name
	 * @param album_art album art file
	 *  
	 */
	public signal void song_changed(string? song, string? artist, string? album, string? album_art);
}

/**
 * JSApiPlayer uses JavaScript API calls to control playback.
 */
public class WebPlayer : Player{
	private JSApi? js_api = null;
	private weak Configuration config;
	private Diorite.Storage storage;
	private Diorite.Connection connection;
	private File? album_art_file = null;
	private int8 album_art_number = -1;
	private WebBackend backend;
	
	[Description(nick = "URL of album art", blurb = "URL of album art")]
	private string? album_art_url {get; private set;} 
	
	
	/**
	 * Creates new JSApiPlayer object.
	 * 
	 * @param config		configuration
	 * @param storage		file storage for album art
	 * @param connection	network connection to download album art
	 * @param actions		user interface actions
	 */
	public WebPlayer(WebBackend backend, Configuration config, Storage storage, Diorite.Connection connection, Diorite.Actions? actions, Diorite.ActionGroups? action_groups)
	{
		base(actions, action_groups);
		this.backend = backend;
		this.config = config;
		this.connection = connection;
		this.storage = storage;
		
		backend.notify["api"].connect_after(on_api_changed);
		if(actions != null){
			actions.get_action(Actions.TOGGLE_PLAY).keybinding = get_pp_accel();
			config.notify["space-key"].connect_after(() => {
				actions.get_action(Actions.TOGGLE_PLAY).keybinding = get_pp_accel();
			});
		}
	}
	
	private void on_api_changed(GLib.Object o, ParamSpec p)
	{
		if (js_api != null)
		{
			js_api.song_changed.disconnect(on_song_changed);
			if (playback_state == STATE_PLAYING)
				pause();
		}
		
		js_api = backend.api;
		song_changed(null, null, null, null);
		playback_state = STATE_NONE;
		if (actions != null)
		{
			actions.get_action(Actions.PREV_SONG).sensitive = false;
			actions.get_action(Actions.NEXT_SONG).sensitive = false;
			actions.get_action(Actions.THUMBS_DOWN).sensitive = false;
			actions.get_action(Actions.THUMBS_UP).sensitive = false;
			actions.get_action(Actions.FAVORITE).sensitive = false;
		}
		
		if (js_api != null)
			js_api.song_changed.connect(on_song_changed);
	}
	
	/**
	 * Response to frame events - updates current status, stores album art
	 * and emit associated signals
	 */
	private void on_song_changed(string? song, string? artist, string? album, string? art_uri,
	string playback_state)
	{
		// Arguments can be null, but not empty.
		return_if_fail(song != "");
		return_if_fail(artist != "");
		return_if_fail(album != "");
		return_if_fail(art_uri != "");
		return_if_fail(playback_state != "");
		
		if(this.playback_state != playback_state){
			debug("Playback state changed: %s → %s",
				this.playback_state ?? "null", playback_state ?? "null");
			this.playback_state = playback_state;
			
		}
		
		var song_changed = false;
		if(this.album_art_url != art_uri){
			debug("Album art changed: %s → %s",
				this.album_art_url ?? "null", art_uri ?? "null");
			this.album_art_url = art_uri;
			if(album_art_file != null){
				try{
					album_art_file.delete();
				}
				catch(Error e){
					warning("Unable to delete old album art file: %s", e.message);
				}
			}
			
			/* Some clients don't reload album art if it has same name, so we reuse file names
			 * album_art.0 - album_art.9. Album art file is not deleted, when Nuvola Player exits,
			 * but the file will be reused in the next run. */
			album_art_number = album_art_number >= 9 ? 0 : album_art_number + 1;
			album_art_file = storage.get_cache_path("album_art." + album_art_number.to_string());
			
			if(art_uri != null && connection.download_file(art_uri, album_art_file)){
				this.album_art = album_art_file.get_path();
			}
			else{
				this.album_art = null;
			}
			song_changed = true;
		}
		
		if(this.artist != artist){
			debug("Artist changed: %s → %s",
				this.artist ?? "null", artist ?? "null");
			this.artist = artist;
			song_changed = true;
		}
		
		if(this.song != song){
			debug("Song changed: %s → %s",
				this.song ?? "null", song ?? "null");
			this.song = song;
			song_changed = true;
		}
		
		if(this.album != album){
			debug("Album changed: %s → %s",
				this.album ?? "null", album ?? "null");
			this.album = album;
			song_changed = true;
		}
		
		if(song_changed){
			this.song_changed(song, artist, album, album_art);
		}
	}
	
	/**
	 * Trigger player action
	 * 
	 * @param action action
	 * @param notify if true a notification will be shown if needed
	 */
	private void trigger_action(string action){
		return_if_fail(js_api != null);
		try{
			this.js_api.send_message(action);
		}
		catch(JSError e){
			critical("Player: %s", e.message);
		}
	}
	
	
	/**
	 * Skips to the next track in the tracklist.
	 * 
	 * If there is no next track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void next_song(){
		this.trigger_action(Actions.NEXT_SONG);
	}
	
	/***
	 * Skips to the previous track in the tracklist.
	 * 
	 * If there is no previous track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void previous_song(){
		this.trigger_action(Actions.PREV_SONG);
	}
	
	/**
	 * Pauses playback.
	 * 
	 * If playback is already paused, this has no effect. Calling Play after this should
	 * cause playback to start again from the same position.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void pause(){
		this.trigger_action(Actions.PAUSE);
	}
	
	/**
	 * Pauses playback.
	 * If playback is already paused, resumes playback.
	 * If playback is stopped, starts playback.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void toggle_play(){
		this.trigger_action(Actions.TOGGLE_PLAY);
	}
	
	/**
	 * Stops playback.
	 * 
	 * If playback is already stopped, this has no effect. Calling Play after this should
	 * cause playback to start again from the beginning of the track. 
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void stop(){
		trigger_action(Actions.PAUSE);
	}
	
	/**
	 * Starts or resumes playback.
	 * 
	 * If already playing, this has no effect.
	 * If there is no track to play, this has no effect.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void play(){
		this.trigger_action(Actions.PLAY);
	}
	
	/**
	 * Sets Thumbs Up (= awesome) to song
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void thumbs_up(){
		this.trigger_action(Actions.THUMBS_UP);
	}
	
	/**
	 * Sets Thumbs Dowb (= horrible) to song
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void thumbs_down(){
		this.trigger_action(Actions.THUMBS_DOWN);
	}
	
	/**
	 * Marks song as favorited
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public override void favorite(){
		this.trigger_action(Actions.FAVORITE);
	}
	
	private string get_pp_accel(){
		return config.space_key ? "<ctrl>space" : "<ctrl>P";
	}
	
}

} // namespace Nuvola
