/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using WebKit;
using Soup;
using Gee;
using Diorite;
using Diorite.Logger;
using Diorite.Widgets;
using Diorite.System;
using JS;
using Nuvola.JSTools;

namespace Nuvola
{

namespace Actions
{
	public const string GO_HOME = "home";
	public const string RELOAD = "refresh";
	public const string GO_BACK = "back";
	public const string GO_FORWARD = "forward";
	public const string ZOOM_IN = "zoom-reset";
	public const string ZOOM_OUT = "zoom-out";
	public const string ZOOM_RESET = "zoom-in";
}

/**
 * Frame with the web interface of cloud player service.
 * Handles state changes received from WebView and produced by script
 * loaded in WebView, console logging, WebView requests, user scripts management, etc.
 */
public class MainWebView: Nuvola.WebView{
	private weak Diorite.Actions? actions;
	private Diorite.Storage storage;
	private Service? _service = null;
	public bool log_navigation = false;
	public Service? service{
		get{
			return _service;
		}
		set{
			set_maintains_back_forward_list(false);
			set_maintains_back_forward_list(true);
			_service = value;
		}
	}
	
	private bool _actions_enabled = true;
	public bool actions_enabled
	{
		get
		{
			return _actions_enabled;
		}
		set
		{
			if (_actions_enabled == value)
				return;
			_actions_enabled = value;
			update_actions();
		}
	}
	
	public virtual signal void navigation_approved(string uri){
		debug_navigation("Navigation approved: %s", uri);
	}
	
	/**
	 * Constructs new Frame for application
	 * 
	 * @param app Application object
	 */
	public class MainWebView(Diorite.Storage storage, Diorite.Actions? actions=null){
		base();
		this.actions = actions;
		this.storage = storage;
		append_ui_actions();
		var settings =  this.get_settings();
		settings.enable_default_context_menu = true;
		this.set_settings(settings);

		navigation_policy_decision_requested.connect(on_new_page);
		new_window_policy_decision_requested.connect(on_new_window);
		
		button_release_event.connect(on_mouse_button_released);
	}
	
	public override WebKit.WebView create_web_view(WebFrame frame){
		/* Create new popup window */
		var popup = new PopupWebView(this.get_settings());
		if (developer_mode)
			popup.developer_mode = true;
		return popup;
	}
	
	/**
	 * Loads home page (see home_page property)
	 */
	public void go_home(){
		if(service != null && service.home_page != null){
			open(service.home_page);
		}
	}
	
	private void on_load_status_changed(GLib.Object o, ParamSpec p)
	{
		if (this._actions_enabled)
		{
			actions.get_action(Actions.GO_BACK).sensitive = can_go_back();
			actions.get_action(Actions.GO_FORWARD).sensitive = can_go_forward();
		}
	}
	
	private void update_actions()
	{
		return_if_fail(actions != null);
		
		string[] names = {Actions.RELOAD, Actions.GO_HOME, Actions.ZOOM_IN, Actions.ZOOM_OUT, Actions.ZOOM_RESET};
		foreach(string name in names)
			actions.get_action(name).sensitive = _actions_enabled;
		
		actions.get_action(Actions.GO_BACK).sensitive = _actions_enabled && can_go_back();
		actions.get_action(Actions.GO_FORWARD).sensitive = _actions_enabled && can_go_forward();
	}
	
	/**
	 * Creates UI actions and appends them to UIActions instance
	 */
	private void append_ui_actions(){
		if(actions == null) return;
	
		/* Normal items. Name, stock id, label, accelerator, tooltip, callback */
		
		actions.add_actions({
			/// Tool bar item tooltip
			new NormalAction(Actions.GO_HOME, Gtk.Stock.HOME, null, "<alt>Home", _("Back to main page of the service"), go_home),
			/// Tool bar item tooltip
			new NormalAction(Actions.RELOAD, Gtk.Stock.REFRESH, null, "F5", _("Reload current page"), reload),
			/// Tool bar item tooltip
			new NormalAction(Actions.GO_BACK, Gtk.Stock.GO_BACK, null, "<shift><alt>Left", _("Go to previous page"), go_back),
			/// Tool bar item tooltip
			new NormalAction(Actions.GO_FORWARD, Gtk.Stock.GO_FORWARD, null, "<shift><alt>Right", _("Go to next page"), go_forward),
			new NormalAction(Actions.ZOOM_RESET, Gtk.Stock.ZOOM_100, null, "<ctrl>0", null, zoom_reset),
			new NormalAction(Actions.ZOOM_IN, Gtk.Stock.ZOOM_IN, null, "<ctrl>plus", null, zoom_in),
			new NormalAction(Actions.ZOOM_OUT, Gtk.Stock.ZOOM_OUT, null, "<ctrl>minus", null, zoom_out)
		});
		
		update_actions();
		
		notify["load-status"].connect(on_load_status_changed);
		notify["uri"].connect(on_load_status_changed);
		notify["title"].connect(on_load_status_changed);
	}
	
	/**
	 * Decide whether open page in main view or in the default browser
	 */
	private bool on_new_page(WebFrame frame, NetworkRequest req,
		WebNavigationAction action, WebPolicyDecision decision
	){
		if(frame != get_main_frame()){
			// don't care about pages in <iframe> elements
			return false;
		}
		
		string uri = req.get_uri();
		debug_navigation("New page request: URI: %s, frame URI: %s",
			uri, frame.get_uri() ?? "null");
			
		
		/* Open Flash install page in default browser */
		if(uri.has_prefix("http://get.adobe.com/flashplayer")){
			open_uri_with_fallback(uri);
			decision.ignore();
			return true;
		}
		
		if(service != null){
			// If page doesn't match sandbox, open it in default browser
			if(!service.matches_sandbox(uri)){
				open_uri_with_fallback(uri);
				decision.ignore();
				return true;
			}
			else{
				navigation_approved(uri);
			}
		}
		return false;
	}
	
	/**
	 * Open new window in the default web browser
	 */
	private bool on_new_window(WebFrame frame, NetworkRequest req,
		WebNavigationAction action, WebPolicyDecision decision
	){
		unowned string uri = req.get_uri();
		debug_navigation("New window request: %s", uri);
		
		if(service != null && service.matches_sandbox(uri)){
			// Oh wait! This page belongs to the main wiew!
			open(uri);
		}
		else{
			open_uri_with_fallback(uri);
		}
		decision.ignore();
		return true;
	}
	
	/**
	 * Handles special mouse buttons (back & forward navigation)
	 */
	private bool on_mouse_button_released(Gdk.EventButton button)
	{
		switch (button.button)
		{
			case 8:  // mouse back button
				go_back();
				return true;
			case 9:  // mouse forward button
				go_forward();
				return true;
			default: // ignore all other buttons
				return false;
		}
	}
	
	/**
	 * Logs debug message with libfenryxo domain // FIXME
	 * @param format format string
	 */
	private void debug_navigation(string format, ...){
		if (log_navigation)
		{
			va_list args = va_list();
			GLib.logv(Config.LOG_DOMAIN, LogLevelFlags.LEVEL_DEBUG, format, args);
		}
	}
}

} // namespace Nuvola
