/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;
using Diorite.Widgets;
using Diorite.System;

namespace Nuvola{
	
public enum UIActionsMode{
	BOTH,
	MENUBAR,
	TOOLBAR;
}

public class MainWindow : Gtk.Window{
	private static const int MINIMAL_REMEMBERED_WINDOW_SIZE = 300;
	private weak Diorite.Application app;
	private Diorite.MultiTypeMap? ui_state;
	private Gtk.UIManager ui_manager;
	private weak Diorite.Actions actions;
	private Diorite.Storage storage;
	private uint ui_merge_id = 0;
	
	// TODO: make private, currently used for sizing kung-fu in extensions/lyrics.vala
	public Gtk.Paned paned{ get; private set;}
	public Gtk.Box box { get; private set;}
	public Gtk.ScrolledWindow content{ get; private set;}
	public Gtk.Overlay overlay{ get; private set;}
	
	private weak Gtk.MenuBar? menubar;

	public signal void close_action();
	
	/**
	 * Constructs new main window
	 * 
	 * @param app Application object
	 */
	public class MainWindow(Diorite.Application app, Gtk.UIManager ui_manager,
		Diorite.Actions actions, Diorite.Storage storage, Diorite.MultiTypeMap? ui_state=null
	){
		this.app = app;
		this.ui_state = ui_state;
		this.ui_manager = ui_manager;
		this.actions = actions;
		this.storage = storage;
		
		window_position = Gtk.WindowPosition.CENTER;
		this.title = app.display_name;
		try {
			this.icon = Gtk.IconTheme.get_default().load_icon(this.app.icon, 48, 0);
		}
		catch(Error e){
			warning("Unable to load application icon.");
		}
		
		// TODO: port to Gtk.Grid
		box = new Gtk.Box(Gtk.Orientation.VERTICAL, 0);
		
		this.content = new Gtk.ScrolledWindow(null, null);
		this.content.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
		
		paned = new Gtk.Paned(Gtk.Orientation.HORIZONTAL);
		paned.pack1(content, true, false);
		
		overlay = new Gtk.Overlay();
		overlay.add(paned);
		overlay.show();
		box.pack_end(overlay);
		base.add(this.box);
		this.box.show_all();

		if(ui_state != null){
			/* Restore last window state */
			int w = 0;
			int h = 0;
			var opt_w = ui_state["window_w"];
			if(opt_w != null) w = int.parse(opt_w);
			var opt_h = ui_state["window_h"];
			if(opt_h != null) h = int.parse(opt_h);
			this.set_default_size(w > MINIMAL_REMEMBERED_WINDOW_SIZE ? w: 1010, h > MINIMAL_REMEMBERED_WINDOW_SIZE ? h : 600);
			
			var opt_x = ui_state["window_x"];
			var opt_y = ui_state["window_y"];
			if(opt_x != null && opt_y != null){
				this.move(int.parse(opt_x), int.parse(opt_y));
			}
			
			if(ui_state.get_bool("window_maximized", false)){
				this.maximize();
			}
			
		}
		else
		{
			this.set_default_size(1010, 600);
		}
		
		/* save UI state at exit */
		app.shutdown.connect(on_quit);
		window_state_event.connect(on_window_state_event);
		add_accel_group(ui_manager.get_accel_group());
	}
	
	
	/**
	 * Emmited when window is about to be closed.
	 * Default handler hides window, when state is "playing" and quit
	 * application otherwise
	 * 
	 * @param event delete event information
	 */
	public override bool delete_event(Gdk.EventAny event)
	{
		hide();
		Idle.add(wanna_quit_cb);
		return true;
	}
	
	private bool wanna_quit_cb()
	{
		close_action();
		return false;
	}
	
	public bool on_window_state_event(Gdk.EventWindowState event){
		if(ui_state == null) return true;
		bool m = (event.new_window_state & Gdk.WindowState.MAXIMIZED) != 0;
		ui_state.set_bool("window_maximized", m);
		return true;
	} 
	
	/**
	 * Save UI state and quit application
	 */
	private void on_quit()
	{
		if (ui_state == null)
			return;
		if (ui_state["window_maximized"] != "on")
		{
			var size = get_allocated_width();
			if (size > MINIMAL_REMEMBERED_WINDOW_SIZE)
				ui_state["window_w"] = size.to_string();
			size = get_allocated_height();
			if (size > MINIMAL_REMEMBERED_WINDOW_SIZE)
				ui_state["window_h"] = size.to_string();
		}
	}
	
	/**
	 * Emitted when clients should remove their menubar/toolbar items.
	 */
	public signal void remove_actions_ui(Gtk.UIManager manager);
	
	/**
	 * Emitted when clients shoul add their menubar/toolbar items.
	 */
	public signal void add_actions_ui(Gtk.UIManager manager);
	
	public void load_actions_ui(){
		Gtk.MenuBar? menubar;
		remove_actions_ui(ui_manager);
		
		if(ui_merge_id != 0){
			ui_manager.remove_ui(ui_merge_id);
			ui_merge_id = 0;
		}
		
		string ui_file = "ui/menubar.xml"; 
		
		var file = storage.get_data_file(ui_file);
		if(file == null){
			critical("Unable to find UI description file");
			return;
		}
		try{
			ui_merge_id = ui_manager.add_ui_from_file(file.get_path());
		}
		catch(Error e){
			if(file == null){
				critical("Unable to load UI description file: %s", e.message);
				return;
			}
		}
		
		add_actions_ui(ui_manager);
		menubar = ui_manager.get_widget("/menubar") as Gtk.MenuBar;
		
		
		if(this.menubar != null){
			box.remove(this.menubar);
		}
		
		this.menubar = menubar;
		if(menubar != null){
			box.pack_start(menubar, false);
			menubar.show_all();
		}
	}
	
	public void set_sidebar(Gtk.Widget widget){
		paned.pack2(widget, false, false);
	}
	
	
	/**
	 * Populates popup menu with actions.
	 * 
	 * @param menu		popup menu to populate
	 */
	public void populate_popup_menu(Gtk.Menu menu)
	{
		var ui_file = "ui/popup_menu.xml";
		var file = storage.get_data_file(ui_file);
		if (file == null)
		{
			critical("Unable to find file '%s'.", ui_file);
			return;
		}
		
		var parser = new Diorite.SimpleUIParser(actions);
		try
		{
			parser.load_from_file(file);
		}
		catch (MarkupError e)
		{
			critical("Parsing of file '%s' failed: %s.", file.get_path(), e.message);
			return;
		}
		
		var orig_menu = parser.get_widget("popupmenu") as Gtk.Container;
		if (orig_menu == null)
		{
			critical("Popup menu not found in file '%s'.", file.get_path());
			return;
		}
		
		foreach (var child in orig_menu.get_children())
		{
			Gtk.MenuItem? item = child as Gtk.MenuItem;
			orig_menu.remove(child);
			menu.append(item);
			item.show();
		}
	}
}

} // namespace Nuvola
