/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * WebView that has modified some web settings.
 */
public class Nuvola.WebView: WebKit.WebView
{
	private const double DEFAULT_DPI = 96.0;
	private double dpi = DEFAULT_DPI;
	private bool _developer_mode = false;
	public bool developer_mode
	{
		get
		{
			return _developer_mode;
		}
		set
		{
			_developer_mode = value;
			var settings =  get_settings();
			settings.enable_default_context_menu = value;
			settings.enable_developer_extras = value;
			set_settings(settings);
			if (value)
				new InspectorWindow(this);
		}
	}
	
	private float _scaled_zoom_level = 1.0f;
	[Description(nick = "DPI-scaled zoom level", blurb = "Zoom level that takes font scaling factor into account.")]
	public float scaled_zoom_level
	{
		get
		{
			if (_scaled_zoom_level != 0.0f)
				return _scaled_zoom_level;
			return (float) (zoom_level * DEFAULT_DPI / dpi);
		}
		set
		{
			if (get_realized())
				zoom_level = (float) (value * dpi / DEFAULT_DPI);
			else
				_scaled_zoom_level = value;
		}
	}
	
	/**
	 * Constructs a new Web View
	 */
	public class WebView()
	{
		var settings =  get_settings();
		settings.enable_caret_browsing = false;
		settings.enable_plugins = false;
		settings.javascript_can_open_windows_automatically = false;
		settings.enable_site_specific_quirks  = true;
		set_settings(settings);
		full_content_zoom = true;
		developer_mode = false;
		style_updated.connect_after(on_style_updated);
	}
	
	public override void realize(){
		update_dpi();
		base.realize();
		if (_scaled_zoom_level != 0.0f)
		{
			scaled_zoom_level = _scaled_zoom_level;
			_scaled_zoom_level = 0.0f;
		}
	}
	
	private void update_dpi()
	{
		weak Gdk.Screen? screen = get_screen();
		if (screen == null)
		{
			warning("WebView.update_dpi() called without a screen.");
			screen = Gdk.Screen.get_default ();
		}
		dpi = screen.get_resolution();
	}
	
	private void on_style_updated(){
		var old_dpi = dpi;
		update_dpi();
		if (old_dpi != dpi)
		{
			zoom_level = (float) (zoom_level * dpi / old_dpi);
		}
	}
	
	/**
	 * Resets zoom level
	 */		
	public void zoom_reset(){
		scaled_zoom_level = 1.0f;
	}
}
